!***************************************************
! One dimensional Split of Jacobi Iteration
! Qun Liu  contact by liu-q14@mails.tsinghua.edu.cn
! Time: 2015-3-27
!***************************************************

program main

    use mpi
    implicit none

    integer, parameter:: totalsize  = 16
    integer, parameter:: mysize     = totalsize/4 
    integer, parameter:: steps      = 100
    integer n, myid, numprocs, i, j, rc
    real a(totalsize, mysize+2),b(totalsize, mysize+2)

    integer begin_col, end_col, ierr
    integer left, right, tag1, tag2
    integer status(MPI_STATUS_SIZE)
    
    ! mpi Initialization
    call MPI_INIT(ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD, myid, ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, numprocs, ierr)
    
    ! Array Initialization
    do j=1,mysize+2
        do i=1, totalsize
            a(i,j)=0.0
        end do
    end do

    if (myid==0) then
        do i=1,totalsize
            a(i,2)=8.0
        end do
    end if

    if (myid==3) then
        do i=1,totalsize
            a(i,mysize+1)=8.0
        end do
    end if

    do j=2, mysize+1 !2:mysize+1, ERROR
        a(1, j)=8.0
        a(totalsize, j)=8.0
    end do

    ! Set left and right process ID
    if (myid > 0) then
        left = myid-1
    else
        left = MPI_PROC_NULL
    end if
    if (myid < 3) then
        right = myid+1
    else
        right = MPI_PROC_NULL
    end if

    ! Communication among processes
    tag1 = 1
    tag2 = 2
    do n=1, steps
        ! data from left to right
        !!!! NOTE: If the line is too long, there would be an error !!!!
        call MPI_SENDRECV(a(1,mysize+1), totalsize, MPI_REAL, right, tag1, &
            a(1,1), totalsize, MPI_REAL, left, tag1, MPI_COMM_WORLD, status, ierr)
        ! data from right to left
        call MPI_SENDRECV(a(1,2), totalsize, MPI_REAL, left, tag2, &
            a(1,mysize+2), totalsize, MPI_REAL, right, tag2, MPI_COMM_WORLD, status, ierr)
        
        ! Do Jacobi iteration
        begin_col = 2
        end_col = mysize+1

        if (myid == 0) then
            begin_col=3
        end if

        if (myid == 3) then
            end_col = mysize
        end if

        do j=begin_col, end_col
            do i=2, totalsize-1
                b(i, j)=(a(i,j+1)+a(i,j-1)+a(i+1,j)+a(i-1,j))*0.25
            end do
        end do

        do j=begin_col, end_col
            do i=2, totalsize-1
                a(i,j)=b(i, j)
            end do
        end do
    end do

    write(*,'(I2,16(4f6.2,/))') myid, ((a(i,j), j=2,mysize+1),i=1,totalsize)

    call MPI_FINALIZE(ierr)

end program main 